from .utils import pre_parse, remove_noise_node, config, html2element, normalize_text
from .extractor import ContentExtractor, TitleExtractor, TimeExtractor, AuthorExtractor, ListExtractor, LangExtractor, \
    HeadMetaExtractor
from newspaper import Config, Article


class GeneralNewsExtractor:
    def __init__(self):
        # 配置 newspaper
        self.config = Config()
        self.article = Article('')

    def extract(self,
                html,
                title_xpath='',
                author_xpath='',
                publish_time_xpath='',
                host='',
                body_xpath='',
                noise_node_list=None,
                with_body_html=False,
                use_visiable_info=False):
        # 对 HTML 进行预处理可能会破坏 HTML 原有的结构，导致根据原始 HTML 编写的 XPath 不可用
        # 因此，如果指定了 title_xpath/author_xpath/publish_time_xpath，那么需要先提取再进行
        self.article.set_html(html)
        self.article.parse()
        # 预处理
        normal_html = normalize_text(html)
        element = html2element(normal_html)
        lang = LangExtractor().language(html)
        headmeta = HeadMetaExtractor().extractor(element)
        title = TitleExtractor().extract(element, title_xpath=title_xpath)
        author = AuthorExtractor().extractor(element, author_xpath=author_xpath) or self.article.authors
        element = pre_parse(element)
        remove_noise_node(element, noise_node_list)
        if not host:
            host = headmeta.get('host')
        content = ContentExtractor().extract(element,
                                             host=host,
                                             with_body_html=with_body_html,
                                             body_xpath=body_xpath,
                                             use_visiable_info=use_visiable_info)
        publish_time = TimeExtractor().extractor(element,
                                                 publish_time_xpath=publish_time_xpath,
                                                 normal_html=normal_html,
                                                 title=title,
                                                 content=content[0][1]['text'],
                                                 html=html,
                                                 npp_pt=self.article.publish_date
                                                 )
        text = content[0][1]['text']
        npp_text = self.article.text
        if npp_text and len(npp_text) > len(text):
            text = npp_text
        result = {'title': title,
                  'author': author,
                  'publish_time': publish_time,
                  'lang': lang,
                  'content': text,
                  'images': content[0][1]['images'],
                  'headmeta': headmeta,
                  'top_image': headmeta.pop('top_image'),
                  'website': headmeta.pop('host')
                  }
        if with_body_html or config.get('with_body_html', False):
            result['body_html'] = content[0][1]['body_html']
        return result


class ListPageExtractor:
    def extract(self, html, feature):
        normalize_html = normalize_text(html)
        element = html2element(normalize_html)
        extractor = ListExtractor()
        return extractor.extract(element, feature)



# -*- coding: utf-8 -*-

import re
from copy import deepcopy

import lxml.html
from lxml.html import HtmlElement
from lxml.html import etree

from ..defaults import USELESS_TAG, TAGS_CAN_BE_REMOVE_IF_EMPTY, USELESS_ATTR, MAYBEIS, INLINE_TAG
from ..defaults import USELESS_TAG_TO_TIME


class DomHandle:

    # 移除 br 和 /*...*/ 注释
    def remove_annotation(self, html):
        html = re.sub('</?br.*?>', '', html, flags=re.S)
        html = re.sub('<!--.*?-->', '', html, flags=re.S)
        html = re.sub(r'/\*.*?\*/', '', html, flags=re.S)
        return html

    # 构建dom树
    def build_html_element(self, html):
        # element = fromstring(html)
        utf8_parser = lxml.html.HTMLParser(encoding='utf-8')
        element = lxml.html.document_fromstring(html.encode('utf-8', 'replace'), parser=utf8_parser)
        return element

    # 移除无用标签
    def remove_useless_tag(self, element: HtmlElement):
        etree.strip_elements(element, *USELESS_TAG, with_tail=False)
        return element

    # 移除无用标签 保留script
    def remove_useless_tag_time(self, element: HtmlElement):
        etree.strip_elements(element, *USELESS_TAG_TO_TIME, with_tail=False)
        return element

    def iter_node(self, element: HtmlElement):
        yield element
        for sub_element in element:
            if isinstance(sub_element, HtmlElement):
                yield from self.iter_node(sub_element)

    # 删除节点
    def remove_node(self, node: HtmlElement):
        parent = node.getparent()
        if parent is not None:
            # print('删除的标签2----', bytes.decode(tostring(node, encoding='utf-8', pretty_print=False, method='html')))
            parent.remove(node)

    # 仅删除标签
    def drop_tag(self, node: HtmlElement):
        parent = node.getparent()
        if parent is not None:
            # print('删除的标签----', bytes.decode(tostring(node, encoding='utf-8', pretty_print=False, method='html')))
            node.drop_tag()

    # 判断节点是否为空
    def is_empty_element(self, node: HtmlElement):
        return not node.getchildren() and not node.text

    # 标准化节点
    def normalize_node(self, element: HtmlElement, tag):
        for node in self.iter_node(element):
            # ['section', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'span']  移除为空的标签
            if node.tag.lower() in TAGS_CAN_BE_REMOVE_IF_EMPTY and self.is_empty_element(node):
                self.remove_node(node)
            # 删除标签(不删文本)
            if node.tag.lower() == 'p':
                etree.strip_tags(node, 'span')
                etree.strip_tags(node, 'strong')
                etree.strip_tags(node, 'font')
            # 不包含子节点的div 改成tagdiv标签 疑似正文标签
            # if node.tag.lower() == 'div' and not node.getchildren():
            #     node.tag = 'tagdiv'

            # 不包含子节点的span 改成p标签
            if tag == 'p' and node.tag.lower() == 'span' and not node.getchildren():
                node.tag = 'p'

            # 删除为空的p标签
            if node.tag.lower() == 'p' and not node.xpath('.//*'):
                if not (node.text and node.text.strip()):
                    self.drop_tag(node)
        return element

    def remove_unlikely_content_tag(self, element: HtmlElement):
        for node in self.iter_node(element):
            id_and_class = '{}{}'.format(node.get('class', ''), node.get('id', ''))
            if id_and_class.strip():
                unlike = '|'.join(USELESS_ATTR)
                maybe = '|'.join(MAYBEIS)
                if re.compile(unlike, re.I).search(id_and_class) and (
                        not re.compile(maybe, re.I).search(id_and_class)) and node.tag not in ['html', 'body']:
                    self.remove_node(node)
        return element

    def remove_noise_node(self, element: HtmlElement, noise_xpath_list):
        if not noise_xpath_list:
            return
        for noise_xpath in noise_xpath_list:
            nodes = element.xpath(noise_xpath)
            for node in nodes:
                self.remove_node(node)
        return element

    def get_text(self, html='', element=None, seg=' ', remove_noise=False, body=False):
        if isinstance(element, HtmlElement):
            selector = deepcopy(element)
        elif html and isinstance(html, str):
            selector = self.build_html_element(html)
            if remove_noise:
                etree.strip_elements(selector, *USELESS_TAG)
        if body:
            text_list = selector.xpath('//body//text()')
        else:
            text_list = selector.xpath('.//text()')
        return seg.join([i.strip() for i in text_list]).strip()

    def remove_inline_tag(self, element):
        elem = deepcopy(element)
        for node in self.iter_node(elem):
            tag_name = node.tag.lower()
            if tag_name in INLINE_TAG:
                # etree.strip_tags(parent, tag_name)
                node.drop_tag()
        return elem



import copy

from lxml.html import Element

from ..defaults import ARTICLE_BODY_TAGS
from ..parsers import Parser


class ArticleBodyExtractor:
    def __init__(self, config):
        self.top_node = None
        self.top_node_complemented = None
        self.parser = Parser()
        self.stopwords_class = config.stopwords_class
        self.language = config.language

    def parse(self, doc: Element):
        """_summary_

        Args:
            doc (Element): _description_
        """
        self.top_node = self.calculate_best_node(doc)
        self.top_node_complemented = self.complement_with_siblings(self.top_node)

    def calculate_best_node(self, doc):
        top_node = None
        nodes_to_check = self.nodes_to_check(doc)
        self.boost_highly_likely_nodes(doc)
        starting_boost = float(1.0)
        cnt = 0
        i = 0
        parent_nodes = []
        nodes_with_text = []

        for node in nodes_to_check:
            text_node = self.parser.getText(node)
            word_stats = self.stopwords_class(
                language=self.language
            ).get_stopword_count(text_node)
            high_link_density = self.is_highlink_density(node)
            if word_stats.stop_word_count > 2 and not high_link_density:
                nodes_with_text.append(node)

        nodes_number = len(nodes_with_text)
        negative_scoring = 0
        bottom_negativescore_nodes = float(nodes_number) * 0.25

        for node in nodes_with_text:
            boost_score = float(0)
            # boost
            if self.is_boostable(node):
                if cnt >= 0:
                    boost_score = float((1.0 / starting_boost) * 50)
                    starting_boost += 1
            # nodes_number
            if nodes_number > 15:
                if (nodes_number - i) <= bottom_negativescore_nodes:
                    booster = float(bottom_negativescore_nodes - (nodes_number - i))
                    boost_score = float(-pow(booster, float(2)))
                    negscore = abs(boost_score) + negative_scoring
                    if negscore > 40:
                        boost_score = float(5)

            text_node = self.parser.getText(node)
            word_stats = self.stopwords_class(
                language=self.language
            ).get_stopword_count(text_node)
            upscore = int(word_stats.stop_word_count + boost_score)

            parent_node = self.parser.getParent(node)
            self.update_score(parent_node, upscore)
            self.update_node_count(parent_node, 1)

            if parent_node not in parent_nodes:
                parent_nodes.append(parent_node)

            # Parent of parent node
            parent_parent_node = self.parser.getParent(parent_node)
            if parent_parent_node is not None:
                self.update_node_count(parent_parent_node, 1)
                self.update_score(parent_parent_node, upscore / 2)
                if parent_parent_node not in parent_nodes:
                    parent_nodes.append(parent_parent_node)
            cnt += 1
            i += 1

        if parent_nodes:
            parent_nodes.sort(key=self.get_score, reverse=True)
            top_node = parent_nodes[0]

        return top_node

    def nodes_to_check(self, doc):
        """Returns a list of nodes we want to search
        on like paragraphs and tables
        """
        nodes_to_check = []
        for tag in ["p", "pre", "td", "article", "div"]:
            if tag == "div":
                items = []
                for id_ in ["article-body", "article", "story", "article-content"]:
                    items += self.parser.getElementsByTag(
                        doc, tag=tag, attr="id", value=id_
                    )
            else:
                items = self.parser.getElementsByTag(doc, tag=tag)
            nodes_to_check += items
        return nodes_to_check

    def is_boostable(self, node):
        """A lot of times the first paragraph might be the caption under an image
        so we'll want to make sure if we're going to boost a parent node that
        it should be connected to other paragraphs, at least for the first n
        paragraphs so we'll want to make sure that the next sibling is a
        paragraph and has at least some substantial weight to it.
        """
        para = "p"
        steps_away = 0
        minimum_stopword_count = 5
        max_stepsaway_from_node = 3

        nodes = self.walk_siblings(node)
        for current_node in nodes:
            # <p>
            current_node_tag = self.parser.getTag(current_node)
            if current_node_tag == para:
                if steps_away >= max_stepsaway_from_node:
                    return False
                paragraph_text = self.parser.getText(current_node)
                word_stats = self.stopwords_class(
                    language=self.language
                ).get_stopword_count(paragraph_text)
                if word_stats.stop_word_count > minimum_stopword_count:
                    return True
                steps_away += 1
        return False

    def is_highlink_density(self, e):
        """Checks the density of links within a node, if there is a high
        link to text ratio, then the text is less likely to be relevant
        """
        links = self.parser.getElementsByTag(e, tag="a")
        if not links:
            return False

        text = self.parser.getText(e)
        words = [word for word in text.split() if word.isalnum()]
        if not words:
            return True
        words_number = float(len(words))
        sb = []
        for link in links:
            sb.append(self.parser.getText(link))

        link_text = "".join(sb)
        link_words = link_text.split()
        num_link_words = float(len(link_words))
        num_links = float(len(links))
        link_divisor = float(num_link_words / words_number)
        score = float(link_divisor * num_links)
        if score >= 1.0:
            return True
        return False
        # return True if score > 1.0 else False

    def boost_highly_likely_nodes(self, doc: Element):
        """Set a bias score for all nodes under most likely
        article containers. This way we can find articles
        that have little text.

        Args:
            doc (Element): Document to be checked
        """
        candidates = []
        for tag in ["p", "pre", "td", "article", "div"]:
            candidates.extend(self.parser.getElementsByTag(doc, tag=tag))

        for e in candidates:
            if self.is_highly_likly(e):
                for child in e.iterdescendants():
                    self.update_score(child, 25)  # TODO: find an optimum value

    def is_highly_likly(self, node: Element) -> bool:
        """Checks if the node is a well known tag + attributes combination
        for article body containers. This way we can deliver even small
        article bodies with high link density
        see: https://finance.yahoo.com/m/
            0edc9aaa-e3ba-3178-be5a-f8c16fbffff2/warren-buffett-stocks-shaky.html

        Args:
            node (Element): Node to check

        Returns:
            bool: True if node could be an article body top node
        """

        def is_tag_match(node, tag_dict):
            if node.tag != tag_dict.get("tag", node.tag):
                return False
            for k, v in tag_dict.items():
                if k == "tag":
                    continue
                if node.get(k) != v:
                    return False
            return True

        for tag in ARTICLE_BODY_TAGS:
            if is_tag_match(node, tag):
                return True

        return False

    def update_score(self, node, add_to_score):
        """Adds a score to the gravityScore Attribute we put on divs
        we'll get the current score then add the score we're passing
        in to the current.
        """
        current_score = 0
        score_string = self.parser.getAttribute(node, "gravityScore")
        if score_string:
            current_score = float(score_string)

        new_score = current_score + add_to_score
        self.parser.setAttribute(node, "gravityScore", str(new_score))

    def update_node_count(self, node, add_to_count):
        """Stores how many decent nodes are under a parent node"""
        current_score = 0
        count_string = self.parser.getAttribute(node, "gravityNodes")
        if count_string:
            current_score = int(count_string)

        new_score = current_score + add_to_count
        self.parser.setAttribute(node, "gravityNodes", str(new_score))

    def add_siblings(self, top_node):
        res_node = copy.deepcopy(top_node)
        baseline_score_siblings_para = self.get_siblings_score(top_node)
        results = self.walk_siblings(top_node)
        for current_node in results:
            ps = self.get_siblings_content(current_node, baseline_score_siblings_para)
            for p in ps:
                res_node.insert(0, p)
        return res_node

    def get_siblings_content(self, current_sibling, baseline_score_siblings_para):
        """Adds any siblings that may have a decent score to this node"""
        if current_sibling.tag == "p" and len(self.parser.getText(current_sibling)) > 0:
            e0 = current_sibling
            if e0.tail:
                e0 = copy.deepcopy(e0)
                e0.tail = ""
            return [e0]
        else:
            potential_paragraphs = self.parser.getElementsByTag(
                current_sibling, tag="p"
            )
            if potential_paragraphs is None:
                return None
            else:
                ps = []
                for first_paragraph in potential_paragraphs:
                    text = self.parser.getText(first_paragraph)
                    if len(text) > 0:
                        word_stats = self.stopwords_class(
                            language=self.language
                        ).get_stopword_count(text)
                        paragraph_score = word_stats.stop_word_count
                        sibling_baseline_score = float(0.30)
                        high_link_density = self.is_highlink_density(first_paragraph)
                        score = float(
                            baseline_score_siblings_para * sibling_baseline_score
                        )
                        if score < paragraph_score and not high_link_density:
                            p = self.parser.createElement(tag="p", text=text, tail=None)
                            ps.append(p)
                return ps

    def get_siblings_score(self, top_node):
        """We could have long articles that have tons of paragraphs
        so if we tried to calculate the base score against
        the total text score of those paragraphs it would be unfair.
        So we need to normalize the score based on the average scoring
        of the paragraphs within the top node.
        For example if our total score of 10 paragraphs was 1000
        but each had an average value of 100 then 100 should be our base.
        """
        base = 100000
        paragraphs_number = 0
        paragraphs_score = 0
        nodes_to_check = self.parser.getElementsByTag(top_node, tag="p")

        for node in nodes_to_check:
            text_node = self.parser.getText(node)
            word_stats = self.stopwords_class(
                language=self.language
            ).get_stopword_count(text_node)
            high_link_density = self.is_highlink_density(node)
            if word_stats.stop_word_count > 2 and not high_link_density:
                paragraphs_number += 1
                paragraphs_score += word_stats.stop_word_count

        if paragraphs_number > 0:
            base = paragraphs_score / paragraphs_number

        return base

    def walk_siblings(self, node):
        return self.parser.previousSiblings(node)

    def get_node_gravity_score(self, node):
        gravity_score = self.parser.getAttribute(node, "gravityScore")
        if not gravity_score:
            return None
        return float(gravity_score)

    def get_score(self, node):
        """Returns the gravityScore as an integer from this node"""
        return self.get_node_gravity_score(node) or 0

    def complement_with_siblings(self, node: Element) -> Element:
        """Adds surrounding relevant siblings to the top node.
        Attention, it generates off-the-tree node.

        Args:
            node (Element): Top node detected

        Returns:
            Element: off the tree node complemented with siblings
        """
        if node is None:
            return node
        node_complemented = self.add_siblings(
            node
        )  # TODO: test if there is a problem with siblings AFTER the top node
        for e in self.parser.getChildren(node_complemented):
            e_tag = self.parser.getTag(e)
            if e_tag != "p":
                if self.is_highlink_density(e):
                    self.parser.remove(e)
        return node_complemented



import re
from copy import deepcopy
from typing import List, Union
from collections import OrderedDict

from lxml.html import HtmlElement

from ..utils import config
from ..parsers import Parser
from ..defaults import AUTHOR_PATTERN, AUTHOR_ATTRS, AUTHOR_STOP_WORDS, AUTHOR_VALS


class AuthorExtractor:
    def __init__(self):
        self.author_pattern = AUTHOR_PATTERN
        self.parser = Parser()
        self.authors: List[str] = []

    def extractor(self, element: HtmlElement, author_xpath='') -> List[str]:
        # 如果提供了自定义的作者 XPath，则直接使用该 XPath 进行提取
        author_xpath = author_xpath or config.get('author', {}).get('xpath')
        if author_xpath:
            self.authors = element.xpath(author_xpath)
        # 否则，从文本中提取作者信息
        text = ''.join(element.xpath('.//text()'))
        for pattern in self.author_pattern:
            author_obj = re.search(pattern, text)
            if author_obj:
                self.authors = self.authors + [author_obj.group(0)]
        self.authors = [i.strip() for i in self.authors if i.strip()]
        # 如果未成功提取作者信息，则尝试从 JSON-LD 数据和其他元素中提取
        if not self.authors:
            _digits = re.compile(r"\d")
            author_stopwords_patt = [re.escape(x) for x in AUTHOR_STOP_WORDS]
            author_stopwords = re.compile(
                r"\b(" + "|".join(author_stopwords_patt) + r")\b", flags=re.IGNORECASE
            )

            def contains_digits(d):
                return bool(_digits.search(d))

            def uniqify_list(lst: List[str]) -> List[str]:
                """
                从提供的字符串列表中去除重复项但保持原始顺序。
                忽略末尾空格和大小写。

                Args:
                    lst (List[str]): 输入的字符串列表，可能包含重复项

                Returns:
                    List[str]: 输出的字符串列表，已删除重复项
                """
                seen = OrderedDict()
                for item in lst:
                    seen[item.lower().strip()] = item.strip()
                return [seen[item] for item in seen.keys() if item]

            def parse_byline(search_str):
                """
                解析候选行的 HTML 或文本，并以列表形式提取出名字
                """
                # 移除 HTML 标签
                search_str = re.sub("<[^<]+?>", "", search_str)
                search_str = re.sub("[\n\t\r\xa0]", " ", search_str)

                # 移除原始的 "By" 语句
                m = re.search(r"\b(by|from)[:\s](.*)", search_str, flags=re.IGNORECASE)
                if m:
                    search_str = m.group(2)

                search_str = search_str.strip()

                # 使用非字母数字字符分割行
                name_tokens = re.split(r"[·,\|]|\sand\s|\set\s|\sund\s|/", search_str)
                # 一些合理性检查
                name_tokens = [s.strip() for s in name_tokens if not contains_digits(s)]
                name_tokens = [s for s in name_tokens if 5 > len(re.findall(r"\w+", s)) > 1]

                return name_tokens

            # Try 1: Search popular author tags for authors

            matches = []
            authors = []

            json_ld_scripts = self.parser.get_ld_json_object(element)

            def get_authors(vals):
                if isinstance(vals, dict):
                    if isinstance(vals.get("name"), str):
                        authors.append(vals.get("name"))
                    elif isinstance(vals.get("name"), list):
                        authors.extend(vals.get("name"))
                elif isinstance(vals, list):
                    for val in vals:
                        if isinstance(val, dict):
                            authors.append(val.get("name"))
                        elif isinstance(val, str):
                            authors.append(val)
                elif isinstance(vals, str):
                    authors.append(vals)

            for script_tag in json_ld_scripts:
                if "@graph" in script_tag:
                    g = script_tag.get("@graph", [])
                    for item in g:
                        if not isinstance(item, dict):
                            continue
                        if item.get("@type") == "Person":
                            authors.append(item.get("name"))
                        if "author" in item:
                            get_authors(item["author"])
                else:
                    if "author" in script_tag:
                        get_authors(script_tag["author"])

            def get_text_from_element(node: HtmlElement) -> str:
                """
                从元素中提取文本，包括其子元素的文本
                """
                if node is None:
                    return ""
                if node.tag in ["script", "style", "time"]:
                    return ""

                node = deepcopy(node)
                for tag in ["script", "style", "time"]:
                    for el in node.xpath(f".//{tag}"):
                        el.getparent().remove(el)
                text = list(node.itertext())
                text = " ".join(text)
                return text

            authors = [re.sub("[\n\t\r\xa0]", " ", x) for x in authors if x]
            doc_root = element.getroottree()

            def getpath(node):
                if doc_root is not None:
                    return doc_root.getpath(node)

            # TODO: be more specific, not a combination of all attributes and values
            for attr in AUTHOR_ATTRS:
                for val in AUTHOR_VALS:
                    found = self.parser.get_element_by_attribs(element, attribs={attr: val})
                    matches.extend([(found, getpath(found)) for found in found])

            matches.sort(
                key=lambda x: x[1], reverse=True
            )  # 按 XPath 排序，以获取最特定的匹配项
            matches_reduced = []
            for m in matches:
                if len(matches_reduced) == 0:
                    matches_reduced.append(m)
                elif not matches_reduced[-1][1].startswith(
                        m[1]
                ):  # 移除前一个节点的父节点
                    matches_reduced.append(m)
            matches_reduced.sort(
                key=lambda x: x[1]
            )  # 保留一些作者的排序

            for match, _ in matches_reduced:
                content: Union[str, List] = ""
                if match.tag == "meta":
                    mm = match.xpath("@content")
                    if len(mm) > 0:
                        content = mm[0]
                else:
                    # 忽略 <time> 标签或带有 "on ..." 的标签
                    # 移除 <style> 标签 https://washingtonindependent.com/how-to-apply-for-reseller-permit-in-washington-state/
                    content = get_text_from_element(match)
                if len(content) > 0:
                    authors.extend(parse_byline(content))

            # 输出的字符串列表，已删除重复项
            authors = [re.sub(author_stopwords, "", x).strip(" .,-/") for x in authors]
            self.authors = uniqify_list(authors)
        return self.authors



from .ContentExtractor import ContentExtractor
from .TitleExtractor import TitleExtractor
from .TimeExtractor import TimeExtractor
from .AuthorExtractor import AuthorExtractor
from .ListExtractor import ListExtractor
from .LangExtractor import LangExtractor
from .HeadMetaExtractor import HeadMetaExtractor


import re
import json
import numpy as np
from lxml.html import etree, tostring
from html import unescape
from ..utils import iter_node, pad_host_for_images, config, get_high_weight_keyword_pattern, universal_filter


class ContentExtractor:
    def __init__(self, content_tag='p'):
        """

        :param content_tag: 正文内容在哪个标签里面
        """
        self.content_tag = content_tag
        self.node_info = {}
        self.high_weight_keyword_pattern = get_high_weight_keyword_pattern()
        self.punctuation = set('''！，。？、；：“”‘’《》%（）,.?:;'"!%()''')  # 常见的中英文标点符号
        self.element_text_cache = {}

    def extract(self, selector, host='', body_xpath='', with_body_html=False, use_visiable_info=False):
        body_xpath = body_xpath or config.get('body', {}).get('xpath', '')
        use_visiable_info = use_visiable_info or config.get('use_visiable_info', False)
        if body_xpath:
            body = selector.xpath(body_xpath)[0]
        else:
            body = selector.xpath('//body')[0]
        for node_index, node in enumerate(iter_node(body)):
            if use_visiable_info:
                if not node.attrib.get('is_visiable', True):
                    continue
                coordinate_json = node.attrib.get('coordinate', '{}')
                coordinate = json.loads(coordinate_json)
                if coordinate.get('height', 0) < 150:  # 正文块的高度应该要大于150px
                    continue
            node_hash = hash(node)
            density_info = self.calc_text_density(node)
            text_density = density_info['density']
            ti_text = density_info['ti_text']
            text_tag_count = self.count_text_tag(node, tag='p')
            sbdi = self.calc_sbdi(ti_text, density_info['ti'], density_info['lti'])
            images_list = [
                i.xpath('./@data-src')[0]
                if i.xpath('./@data-src') and ';base64,' in i.xpath('./@src')[0]
                else i.xpath('./@src')[0]
                for i in node.xpath('.//img')
            ]
            host = host or config.get('host', '')
            if host:
                images_list = [pad_host_for_images(host, url) for url in images_list]
            node_info = {'ti': density_info['ti'],
                         'lti': density_info['lti'],
                         'tgi': density_info['tgi'],
                         'ltgi': density_info['ltgi'],
                         'node': node,
                         'node_index': node_index,
                         'density': text_density,
                         'text': ti_text,
                         'images': images_list,
                         'text_tag_count': text_tag_count,
                         'sbdi': sbdi}
            if use_visiable_info:
                node_info['is_visiable'] = node.attrib['is_visiable']
                node_info['coordinate'] = node.attrib.get('coordinate', '')
            if with_body_html or config.get('with_body_html', False):
                body_source_code = unescape(etree.tostring(node, encoding='utf-8').decode())
                node_info['body_html'] = universal_filter(body_source_code)
            self.node_info[node_hash] = node_info
        self.calc_new_score()
        result = sorted(self.node_info.items(), key=lambda x: x[1]['score'], reverse=True)
        return result

    def count_text_tag(self, element, tag='p'):
        """
        当前标签下面的 text()和 p 标签，都应该进行统计
        :param element:
        :param tag:
        :return:
        """
        tag_num = len(element.xpath(f'.//{tag}'))
        direct_text = len(element.xpath('text()'))
        return tag_num + direct_text

    def get_all_text_of_element(self, element_list):
        if not isinstance(element_list, list):
            element_list = [element_list]

        text_list = []
        for element in element_list:
            element_flag = element.getroottree().getpath(element)
            if element_flag in self.element_text_cache:  # 直接读取缓存的数据，而不是再重复提取一次
                text_list.extend(self.element_text_cache[element_flag])
            else:
                element_text_list = []
                for text in element.xpath('.//text()'):
                    text = text.strip()
                    if not text:
                        continue
                    clear_text = re.sub(' +', ' ', text, flags=re.S)
                    element_text_list.append(clear_text.replace('\n', ''))
                self.element_text_cache[element_flag] = element_text_list
                text_list.extend(element_text_list)
        return text_list

    def need_skip_ltgi(self, ti, lti):
        """
        有时候，会出现像维基百科一样，在文字里面加a 标签关键词的情况，例如：

        <div>
        我是正文我是正文我是正文<a href="xxx">关键词1</a>我是正文我是正文我是正文我是正文
        我是正文我是正文我是正文我是正文我是正文<a href="xxx">关键词2</a>我是正文我是正文
        我是正文
        </div>

        在这种情况下，tgi = ltgi = 2，计算公式的分母为0. 为了把这种情况和列表页全是链接的
        情况区分出来，所以要做一下判断。检查节点下面所有 a 标签的超链接中的文本数量与本节点
        下面所有文本数量的比值。如果超链接的文本数量占比极少，那么此时，ltgi 应该忽略
        :param ti: 节点 i 的字符串字数
        :param lti: 节点 i 的带链接的字符串字数
        :return: bool
        """
        if lti == 0:
            return False

        return ti // lti > 10  # 正文的字符数量是链接字符数量的十倍以上

    def calc_text_density(self, element):
        """
        根据公式：

               Ti - LTi
        TDi = -----------
              TGi - LTGi


        Ti:节点 i 的字符串字数
        LTi：节点 i 的带链接的字符串字数
        TGi：节点 i 的标签数
        LTGi：节点 i 的带连接的标签数


        :return:
        """
        ti_text = '\n'.join(self.get_all_text_of_element(element))
        ti = len(ti_text)
        ti = self.increase_tag_weight(ti, element)
        a_tag_list = element.xpath('.//a')

        lti = len(''.join(self.get_all_text_of_element(a_tag_list)))
        tgi = len(element.xpath('.//*'))
        ltgi = len(a_tag_list)
        if (tgi - ltgi) == 0:
            if not self.need_skip_ltgi(ti, lti):
                return {'density': 0, 'ti_text': ti_text, 'ti': ti, 'lti': lti, 'tgi': tgi, 'ltgi': ltgi}
            else:
                ltgi = 0
        density = (ti - lti) / (tgi - ltgi)
        return {'density': density, 'ti_text': ti_text, 'ti': ti, 'lti': lti, 'tgi': tgi, 'ltgi': ltgi}

    def increase_tag_weight(self, ti, element):
        tag_class = element.get('class', '')
        tag_id = element.get('id', '')
        if self.high_weight_keyword_pattern.search(tag_id):
            return 2 * ti
        if self.high_weight_keyword_pattern.search(tag_class):
            return 2 * ti
        return ti

    def calc_sbdi(self, text, ti, lti):
        """
                Ti - LTi
        SbDi = --------------
                 Sbi + 1

        SbDi: 符号密度
        Sbi：符号数量

        :return:
        """
        sbi = self.count_punctuation_num(text)
        sbdi = (ti - lti) / (sbi + 1)
        return sbdi or 1  # sbdi 不能为0，否则会导致求对数时报错。

    def count_punctuation_num(self, text):
        count = 0
        for char in text:
            if char in self.punctuation:
                count += 1
        return count

    def calc_new_score(self):
        """
        score = 1 * ndi * log10(text_tag_count + 2) * log(sbdi)

        1：在论文里面，这里使用的是 log(std)，但是每一个密度都乘以相同的对数，他们的相对大小是不会改变的，所以我们没有必要计算
        ndi：节点 i 的文本密度
        text_tag_count: 正文所在标签数。例如正文在<p></p>标签里面，这里就是 p 标签数，如果正文在<div></div>标签，这里就是 div 标签数
        sbdi：节点 i 的符号密度
        :param std:
        :return:
        """
        for node_hash, node_info in self.node_info.items():
            score = node_info['density'] * np.log10(node_info['text_tag_count'] + 2) * np.log(
                node_info['sbdi'])
            self.node_info[node_hash]['score'] = score

    def xpath_extract_content(self, element, content_xpath='', encoding='utf-8', clear_attrib=True, pretty_print=False,
                              method='html'):
        """
        提取带标签的内容
        :param element:
        :param content_xpath:
        :param encoding:
        :param clear_attrib:
        :param pretty_print:
        :param method:
        :return:
        """
        content = element.xpath(content_xpath)
        news_content = ''
        if content:
            for element_ in content:
                if clear_attrib:
                    element_.attrib.clear()
                news_content += bytes.decode(
                    tostring(element_, encoding=encoding, pretty_print=pretty_print, method=method))
        return news_content
